/** @file Fractals.inl
	@author Jukka Jylnki

	This work is copyrighted material and may NOT be used for any kind of commercial or 
	personal advantage and may NOT be copied or redistributed without prior consent
	of the author(s). 

	@brief Contains implementations for different fractal iteration procedures.
*/

#ifdef SSE2
#include <emmintrin.h>

// A bit of crap, but inline function didn't cut it :(
#define MANDELSSE2(real, imag, creal, cimag) \
		__m128d real2 = _mm_mul_pd(real, real); \
		__m128d imag2 = _mm_mul_pd(imag, imag); \
		__m128d realnew = _mm_sub_pd(real2, imag2); \
		__m128d imagnew = _mm_mul_pd(real, imag); \
		imagnew = _mm_mul_pd(imagnew, c2); \
		real = _mm_add_pd(realnew, creal); \
		imag = _mm_add_pd(imagnew, cimag);

/** An SSE2 -optimized Mandelbrot iterator. Does 6 iterations to two pixels at a time and then checks the results. */
class MandelbrotSSE2
{
public:
	static const inline __forceinline void Iterate(IterationEntry &entry, IterationEntry &entry2)
	{
		entry.iterationCount += 6;
		entry2.iterationCount += 6;

		// Load values
		__m128d c2 = _mm_setr_pd(2.0, 2.0);
		__m128d real = _mm_setr_pd(entry.value.r, entry2.value.r);
		__m128d imag = _mm_setr_pd(entry.value.i, entry2.value.i);
		__m128d creal = _mm_setr_pd(entry.initialValue.r, entry2.initialValue.r);
		__m128d cimag = _mm_setr_pd(entry.initialValue.i, entry2.initialValue.i);

		// Crunchies
		{ MANDELSSE2(real, imag, creal, cimag); }
		{ MANDELSSE2(real, imag, creal, cimag); }
		{ MANDELSSE2(real, imag, creal, cimag); }
		{ MANDELSSE2(real, imag, creal, cimag); }
		{ MANDELSSE2(real, imag, creal, cimag); }
		MANDELSSE2(real, imag, creal, cimag); // Intentionally not in its own scope, our (crap) #define gives the constants for store.

		// Store back
		_mm_storel_pd((double*)&entry.value.r, real);
		_mm_storel_pd((double*)&entry.value.i, imag);
		_mm_storeh_pd((double*)&entry2.value.r, real);
		_mm_storeh_pd((double*)&entry2.value.i, imag);

		// Compare results
		if (entry.value.Magnitude2() >= 4.0) // TODO: Optimize to use real2 and imag2
			entry.iterationCount |= cDiverged;
		if (entry2.value.Magnitude2() >= 4.0) // TODO: Optimize to use real2 and imag2
			entry2.iterationCount |= cDiverged;
	}

	static const int Iterations() { return 6; }
};
#else
class MandelbrotSSE2
{
public:
	static const inline __forceinline void Iterate(IterationEntry &entry, IterationEntry &entry2)
	{
		// No support for SSE2, display a checker pattern
		const Real thickness = 1e-3f;
		if (((fmod(abs(entry.value.r), thickness) < thickness/2) ^ (fmod(abs(entry.value.i), thickness) < thickness/2)) != 0)
			entry.iterationCount = cDiverged | 1;
		else
			entry.iterationCount = cDiverged;
	}

	static const int Iterations() { return 1; }
};
#endif

// Again, to force the compiler "inline" and not let it insert a call.
#define MANDEL(zr, zi, cr, ci) \
	const Real x2 = zr * zr; \
	const Real y2 = zi * zi; \
	const Real xy2 = zr * zi * 2.0; \
	zr = x2 - y2 + cr; \
	zi = xy2 + ci;

// The default Mandelbrot implementation
class Mandelbrot
{
public:
	static const inline __forceinline void Iterate(IterationEntry &entry, IterationEntry &entry2)
	{
		{
			++entry.iterationCount;
			MANDEL(entry.value.r, entry.value.i, entry.initialValue.r, entry.initialValue.i);
			if (x2 + y2 >= 4.0)
				entry.iterationCount |= cDiverged;
		}
		{
			++entry2.iterationCount;
			MANDEL(entry2.value.r, entry2.value.i, entry2.initialValue.r, entry2.initialValue.i);
			if (x2 + y2 >= 4.0)
				entry2.iterationCount |= cDiverged;
		}
	}

	static const int Iterations() { return 1; }
};

// Julia set, an arbitrary constant C.
class Julia
{
public:

	static const inline __forceinline void Iterate(IterationEntry &entry)
	{
		++entry.iterationCount;
		const Real r = entry.value.r*entry.value.r-entry.value.i*entry.value.i + Real(0.2);
		entry.value.i = Real(2.0)*entry.value.r*entry.value.i + Real(0.5);
		entry.value.r = r;

		if (entry.value.Magnitude2() >= 4.0)
				entry.iterationCount |= cDiverged;
	}

	static const inline __forceinline void Iterate(IterationEntry &entry, IterationEntry &entry2)
	{
		Iterate(entry);
		Iterate(entry2);
	}

	static const int Iterations() { return 1; }
};

/// Mandelbrot generalization: z = z^T+c
template<int T>
class MandelbrotN
{
public:
	static const inline __forceinline void Iterate(IterationEntry &entry)
	{
		++entry.iterationCount;
		Complex<Real> z(entry.value.r, entry.value.i);
		Complex<Real> v(1,0);
		for(int n = 0; n < T; n++)
			v = v * z;

		entry.value = v + entry.initialValue;

		if (entry.value.Magnitude2() >= 4.0)
				entry.iterationCount |= cDiverged;
	}

	static const inline __forceinline void Iterate(IterationEntry &entry, IterationEntry &entry2)
	{
		Iterate(entry);
		Iterate(entry2);
	}

	static const int Iterations() { return 1; }
};

/// The Mandelbar set, z = conj(z)^T+c
template<int T>
class MandelbrotConjN
{
public:
	static const inline __forceinline void Iterate(IterationEntry &entry)
	{
		++entry.iterationCount;
		Complex<Real> z(entry.value.r, -entry.value.i);
		Complex<Real> v(1,0);
		for(int n = 0; n < T; n++)
			v = v * z;

		entry.value = v + entry.initialValue;

		if (entry.value.Magnitude2() >= 4.0)
				entry.iterationCount |= cDiverged;
	}

	static const inline __forceinline void Iterate(IterationEntry &entry, IterationEntry &entry2)
	{
		Iterate(entry);
		Iterate(entry2);
	}

	static const int Iterations() { return 1; }
};
