/** @file Mandelbrot.cpp
	@author Jukka Jylnki

	This work is copyrighted material and may NOT be used for any kind of commercial or 
	personal advantage and may NOT be copied or redistributed without prior consent
	of the author(s). 

	@brief This sample directly manipulates a texture surface.

	In this sample, we calculate the Mandelbrot set on the CPU and display the results by filling a texture surface
	and rendering it using the GPU. To manipulate the texture surface, we use the Bitmap class.

	It seems that the sample has derailed from a texture surface access sample to a serious performance optimization
	and fractal generator robustness project. The use of SSE2 and OpenMP2.0 language extension can be controlled with
	these preprocessor flags:

	#define SSE2
	#define OPENMP

	For MSVS2005, these can be controlled by compiling the appropriate solution configurations.
*/


#ifdef OPENMP
#include <omp.h>
#endif

#include "clb/clbLib.h"

using namespace clb;

/// Make a nice typedef to allow for easy alteration of the precision for mandelbrot generation.
typedef long double Real;

/// The iteration parameters for each pixel.
class IterationEntry
{
public:
	Complex<Real> value;
	Complex<Real> initialValue;
	dword iterationCount;
};

/// The MSB is set for the IterationEntry::iterationCount when the point has been found to tend to infinity.
static const dword cDiverged = 0x80000000; 

// Various fractal algorithms
#include "Fractals.inl"

/// These are the supported fractals:
enum FractalType
{
	FMandelbrotSSE2,	///< z=z^2+c, optimized with SSE2 (contains a bit of cheating on the algorithm)
	FMandelbrot,		///< z=z^2+c, the exact version to be able to compare against the SSE2 version for quality & performance
	FJulia,				///< Julia set with arbitrary constant C.
	FMandelbrot3,		///< z=z^3+c
	FMandelbrot4,
	FMandelbrot5,
	FMandelbrot6,
	FMandelbrot7,
	FMandelbrotConj,	///< z=conj(z)^2+c
	FMandelbrotConj3, ///< z=conj(z)^3+c
	FMandelbrotConj4,
	FMandelbrotConj5,
	FMandelbrotConj6,
	FMandelbrotConj7,
};

class MandelbrotApp : public ClbLibApp
{
private:
	// Define a palette that is used to give a nice gradient for the image.
	static const int cPaletteEntries = 4096;
	dword palette[cPaletteEntries];

	static const int cMaxIterationsPerFrame = 100; ///< To disallow updating too many times per frame (would spoil the animation)

	/// The type of the currently drawn fractal.
	FractalType currentFractalType;

	/// This texture is used to render the result.
	Ptr(Texture) mandelbrot;
	static const int cTexX = 60; // Render texture to these coordinates
	static const int cTexY = 40;

	static const int cTextureSize = 512; ///< Texture resolution for the mandelbrot image.
	ArrayPtr(IterationEntry) values; ///< Contains the iteration values for each pixel.

	/// Indexes to the mandelbrotValues -array to define the set of points that currently seem to be converging.
	ArrayPtr(int) convergingPoints;
	int numConvergingPoints; ///< Specifies the active size of the above array.
	
	int iterations; ///< The number of total iterations performed on the array.
	Real center[2]; ///< Determines the point at the complex plane that the center pixel of the texture corresponds to.
	Real scale[2]; ///< Current scale value (or rather, size value)

	bool bDrag; ///< true if mouse dragging is going on.
	bool bUniformScale; ///< If true, a constrain scale[0]==scale[1] will be applied.
	int drag[4]; ///< Saves the mouse drag coordinates.

	// Statistics for coloring:
	dword minIters, maxIters;

	// Some statistics for rendering:
	float iterationsPerSec;

	/// Loads the palette used for coloring the image. 
	void LoadPalette()
	{
		// Loads rainbow.xxx image from file.
		Ptr(Texture) paletteTex = Global::resPool.Create<Texture>("rainbow");

		// Here we use the Bitmap object to read from a texture:
		// Read the first row of the texture and scale the color values to a 1D palette array.
		Bitmap b = paletteTex->BeginDraw(); // BeginDraw() locks the texture for per-pixel access
		for(int i = 0; i < cPaletteEntries; i++)
			palette[i] = b.GetPixel(i * b.Width() / cPaletteEntries, 0); // Read using GetPixel()
		paletteTex->EndDraw(); // EndDraw() signals that we are finished using the texture.
	}

	/// Creates a square-sized 2D texture. Note that there will be no mip levels allocated.
	Ptr(Texture) Create2DTexture(int size)
	{
		TextureDesc desc;
		desc.width = desc.height = size;
		desc.pixelFormat = TexARGB8888;
		desc.type = Texture2D;
		desc.mipLevels = 1;

		return Global::resPool.CreateTexture("mandelbrot", desc);
	}

	/// Allocates space for arrays and resets them to proper initial values
	void ResetMandelbrotArrays(Real ulx, Real uly, Real brx, Real bry)
	{
		const int arrayElems = cTextureSize * cTextureSize;

		values = new IterationEntry[arrayElems];
		memset(values, 0, sizeof(IterationEntry) * arrayElems);

		// Initialize the initial values (constants c) for the calculation.
		for(int y = 0; y < cTextureSize; ++y)
			for(int x = 0; x < cTextureSize; ++x)
			{
				values[y * cTextureSize + x].value =
				values[y * cTextureSize + x].initialValue = 
					Complex<Real>(ulx + Real(brx-ulx) * (Real)x/cTextureSize,
					              uly + Real(bry-uly) * (Real)y/cTextureSize);
			}

		// Recreate and reset the list of currently converging points
		convergingPoints = new int[arrayElems];
		for(int i = 0; i < arrayElems; i++)
			convergingPoints[i] = i;
		numConvergingPoints = arrayElems;

		minIters = 0xFFFFFFFF;
		maxIters = 0;

		iterations = 0;
	}

	FractalType ToFractalType(const std::string &name)
	{
		if (name == "Mandelbrot") return FMandelbrot;
		if (name == "Mandelbrot3") return FMandelbrot3;
		if (name == "Mandelbrot4") return FMandelbrot4;
		if (name == "Mandelbrot5") return FMandelbrot5;
		if (name == "Mandelbrot6") return FMandelbrot6;
		if (name == "Mandelbrot7") return FMandelbrot7;
		if (name == "MandelbrotConj") return FMandelbrotConj;
		if (name == "MandelbrotConj3") return FMandelbrotConj3;
		if (name == "MandelbrotConj4") return FMandelbrotConj4;
		if (name == "MandelbrotConj5") return FMandelbrotConj5;
		if (name == "MandelbrotConj6") return FMandelbrotConj6;
		if (name == "MandelbrotConj7") return FMandelbrotConj7;
		if (name == "Julia") return FJulia;
		if (name == "MandelbrotSSE2") return FMandelbrotSSE2;

		return FMandelbrot;
	}

	/// Manages incoming clbLib window events
	void WindowEvent(Window *wnd, WindowMessage message)
	{
		if (wnd->Name() == "FractalType" && message == WMDropboxSelected)
		{
			Dropbox *t = dynamic_cast<Dropbox*>(wnd);
			assert(t);
			FractalType newFractalType = ToFractalType(t->SelectedItem());
			if (newFractalType != currentFractalType)
			{
				currentFractalType = newFractalType;
				ResetWindow();
				ResetMandelbrotArrays(center[0] - scale[0], center[1] - scale[1], center[0] + scale[0], center[1] + scale[1]);
			}
		}
	}

	/// Performs one or more iterations on the mandelbrot array. Returns the number of iterations done.
	template<typename T>
	int Iterate()
	{
		PROFILE(Iterate);

		const int arrayElems = cTextureSize * cTextureSize;

		int convergedPoints = 0;

		// Do a multithreaded loop over all active pixels on the surface.
#ifdef OPENMP
	#pragma omp parallel for reduction (+: convergedPoints)
#endif
		for(int i = arrayElems-numConvergingPoints; i < arrayElems; i += 2)
		{
			int &pt = convergingPoints[i];
			int &pt2 = (i+1 >= arrayElems) ? pt : convergingPoints[i+1];

			// Two pixels at a time are fed to the iteration routine so it can take advantage of vector parallelization.
			T::Iterate(values[pt], values[pt2]);
		}

		// Loop again (now single-threaded) and determine which pixels fell out of the set.
		int i = arrayElems-1;
		for(int j = numConvergingPoints; j > 0; --j, --i)
		{
			int &pt = convergingPoints[i];
			if ((values[pt].iterationCount & cDiverged) != 0)
			{
				minIters = min(minIters, values[pt].iterationCount & ~cDiverged);
				maxIters = max(maxIters, values[pt].iterationCount & ~cDiverged);
				pt = -1;
				std::swap(pt, convergingPoints[arrayElems-numConvergingPoints]);
				++convergedPoints;
				--numConvergingPoints;
				++i;
			}
		}

		// We have now performed a whole iteration (or several of them, if the iterator had an optimization), update the global counter.
		iterations += T::Iterations();
		return T::Iterations();
	}

	/// Returns a color from the loaded palette.
	dword cTextureSizeToColor(int size)
	{
		size = clamp(size, 0, cPaletteEntries-1);
		return palette[size];
	}

	/** Draws the mandelbrot set to a texture.
		@param ul The coordinates at the complex plane for the upper left corner of the texture.
		@param br The coordinates at the complex plane for the bottom right corner of the texture. 
		@param t The texture to draw to.
		@param iterations The maximum number of iterations to do for each point. */
	void DrawMandelbrot(Texture &t)
	{
		PROFILE(DrawMandelbrot);

		// Lock surface for pixel access.
		Bitmap b = t.BeginDraw();

		if (minIters < maxIters)
			for(int by = b.Height()-1; by >= 0; --by)
				for(int bx = b.Width()-1; bx >= 0; --bx)
				{
					const int i = by * cTextureSize + bx;

					// Calculate the 'size' (==magnitude) of the current point c=x+yi.
					const bool bDiverged = (values[i].iterationCount & cDiverged) != 0;
					const dword color = bDiverged ? 
						cTextureSizeToColor(((values[i].iterationCount & ~cDiverged) - minIters) * cPaletteEntries / (maxIters-minIters)) : 0xFF000000;

					b.PutPixel(bx, by, color);
				}
		else
			b.Clear(0);

		// Texture border rectangle
		b.Rectangle(0, 0, b.Width()-1, b.Height()-1, 0xFFFFFFFF);

		// Drag border rectangle
		if (bDrag)
			b.Rectangle(drag[0], drag[1], drag[2], drag[3], 0xFF8080FF);

		// Release the texture lock. Using b is invalid after this point.
		t.EndDraw();
	}

	/// Sets the window area to starting parameters that show the whole set.
	void ResetWindow()
	{
		center[0] = Real(0);
		center[1] = Real(0);
		scale[0] = scale[1] = Real(2);
	}

	/* Translates the view by the given amount of pixels. The already calculated data is moved about in the buffers
		so that they are not lost. */
	void Translate(int xPixels, int yPixels)
	{
		// Translate the center
		center[0] += xPixels * scale[0] * Real(2.0) / Real(cTextureSize);
		center[1] += yPixels * scale[0] * Real(2.0) / Real(cTextureSize);

		// Complex plane extents
		const Real ulx = center[0] - scale[0];
		const Real uly = center[1] - scale[1];
		const Real brx = center[0] + scale[0];
		const Real bry = center[1] + scale[1];

		// Determine the loop extents:
		int yStart = yPixels < 0 ? cTextureSize-1 : 0;
		int yEnd = yPixels < 0 ? -yPixels-1 : cTextureSize-yPixels;
		int yAdd = yStart < yEnd ? 1 : -1;

		int xStart = xPixels < 0 ? cTextureSize-1 : 0;
		int xEnd = xPixels < 0 ? -xPixels-1 : cTextureSize-xPixels;
		int xAdd = xStart < xEnd ? 1 : -1;

		// Copy offsets
		const int xOfs = xPixels;
		const int yOfs = yPixels;

		// Move existing calculated values
		for(int y = yStart; y != yEnd; y += yAdd)
			for(int x = xStart; x != xEnd; x += xAdd)
				values[y*cTextureSize+x] = values[(y+yOfs)*cTextureSize+x+xOfs];

		// Determine loop extents for resetting the new pixels that need to be calculated.
		yStart = yPixels < 0 ? 0 : cTextureSize-yPixels;
		yEnd = yPixels < 0 ? -yPixels : cTextureSize;
		yAdd = yStart < yEnd ? 1 : -1;

		xStart = xPixels < 0 ? 0 : cTextureSize-xPixels;
		xEnd = xPixels < 0 ? -xPixels : cTextureSize;
		xAdd = xStart < xEnd ? 1 : -1;

		// Loop and set proper initial conditions for each new pixel.
		for(int y = yStart; y != yEnd; y += yAdd)
			for(int x = 0; x < cTextureSize; x++)
			{
				values[y*cTextureSize+x].iterationCount = 0;

				values[y * cTextureSize + x].initialValue =
				           values[y*cTextureSize+x].value =
				Complex<Real>(ulx + Real(brx-ulx) * (Real)x/cTextureSize,
				              uly + Real(bry-uly) * (Real)y/cTextureSize);
			}

		// Loop and set proper initial conditions for each new pixel.
		for(int y = 0; y < cTextureSize; y++)
			for(int x = xStart; x != xEnd; x += xAdd)
			{
				values[y*cTextureSize+x].iterationCount = 0;

				values[y * cTextureSize + x].initialValue =
				           values[y*cTextureSize+x].value =
				Complex<Real>(ulx + Real(brx-ulx) * (Real)x/cTextureSize,
				              uly + Real(bry-uly) * (Real)y/cTextureSize);
			}

		// The intermediate arrays used during calculations now have wrong information, just reset them:
		const int arrayElems = cTextureSize * cTextureSize;
		numConvergingPoints = 0;

		minIters = 0xFFFFFFFF;
		maxIters = 0;
		for(int i = 0; i < arrayElems; i++)
		{
			// Determine if this pixel is a converging one.
			const dword &iterCount = values[i].iterationCount;

			if ((iterCount & cDiverged) == 0)
			{
				convergingPoints[arrayElems-numConvergingPoints-1] = i;
				++numConvergingPoints;
			}
			else // this point has diverged, re-evaluate min & max
			{
				minIters = min(minIters, iterCount & ~cDiverged);
				maxIters = max(maxIters, iterCount & ~cDiverged);
			}
		}

		for(int i = 0; i < arrayElems-numConvergingPoints; ++i)
			convergingPoints[i] = -1;
	}

	/** Calls the proper iteration function for one or more iterations over the pixel array.
		@return The number of iterations performed on the array. */
	int IterateFractal()
	{
		switch(currentFractalType)
		{
		case FMandelbrot: return Iterate<Mandelbrot>();
		case FMandelbrotSSE2: return Iterate<MandelbrotSSE2>();
		case FJulia: return Iterate<Julia>();
		case FMandelbrot3: return Iterate<MandelbrotN<3> >();
		case FMandelbrot4: return Iterate<MandelbrotN<4> >();
		case FMandelbrot5: return Iterate<MandelbrotN<5> >();
		case FMandelbrot6: return Iterate<MandelbrotN<6> >();
		case FMandelbrot7: return Iterate<MandelbrotN<7> >();
		case FMandelbrotConj: return Iterate<MandelbrotConjN<2> >();
		case FMandelbrotConj3: return Iterate<MandelbrotConjN<3> >();
		case FMandelbrotConj4: return Iterate<MandelbrotConjN<4> >();
		case FMandelbrotConj5: return Iterate<MandelbrotConjN<5> >();
		case FMandelbrotConj6: return Iterate<MandelbrotConjN<6> >();
		case FMandelbrotConj7: return Iterate<MandelbrotConjN<7> >();
		default:
			FAIL("Unsupported fractal type!");
			return 0;
		}
	}
public:

	/// Called at startup. Initializes all necessary starting data.
	void Init(int, char **)
	{
		LoadPalette();
		mandelbrot = Create2DTexture(cTextureSize);
		ResetWindow();
		ResetMandelbrotArrays(center[0] - scale[0], center[1] - scale[1], center[0] + scale[0], center[1] + scale[1]);
		bDrag = false;
		bUniformScale = true;
		currentFractalType = FMandelbrot;

		Global::resPool.LoadWindows("data/menus.txt");
	}

	// Looks out for input and updates the Mandelbrot iteration appropriate times.
	void FrameMove(float fTimePassed)
	{
		using namespace clb::Global;

		bool bChanged = false;

		// Arrow keys translate complex plane center point
		{
			PROFILE(Translate);
			int tX = 0;
			int tY = 0;
			if (input.IsKeyDown(KeyUp)) tY -= 5;
			if (input.IsKeyDown(KeyDown)) tY += 5;
			if (input.IsKeyDown(KeyLeft)) tX -= 5;
			if (input.IsKeyDown(KeyRight)) tX += 5;

			if (tX != 0 || tY != 0)
				Translate(tX, tY);
		}

		// Reset window
		if (input.IsKeyDown(DIK_R))
		{
			ResetWindow();
			bChanged = true;
		}

		// Mouse Right click zooms out
		if (input.Get(MouseRight) && !input.Get(MouseOldRight) && !bDrag)
		{
			scale[0] *= 2.0;
			scale[1] *= 2.0;

			bChanged = true;
		}

		// Mouse left starts drag
		if (input.Get(MouseLeft) && !input.Get(MouseOldLeft))
		{
			bDrag = true;
			drag[0] = input.Get(MouseX) - cTexX;
			drag[1] = input.Get(MouseY) - cTexY;
		}

		// Mouse right cancels drag
		if (bDrag && input.Get(MouseRight))
			bDrag = false;

		// Update drag coordinates
		if (bDrag)
		{
			drag[2] = input.Get(MouseX) - cTexX;
			drag[3] = input.Get(MouseY) - cTexY;

			// Constrain uniform scale
			if (bUniformScale)
			{
				int uniformScale = max(abs(drag[2]-drag[0]),abs(drag[3]-drag[1]));
				if (drag[2] > drag[0])
					drag[2] = drag[0] + uniformScale;
				else
					drag[2] = drag[0] - uniformScale;
				if (drag[3] > drag[1])
					drag[3] = drag[1] + uniformScale;
				else
					drag[3] = drag[1] - uniformScale;
			}
		}
		// Mouse left release ends drag
		if (!input.Get(MouseLeft) && input.Get(MouseOldLeft) && bDrag)
			if (abs(drag[2]-drag[0]) > 1 && abs(drag[3]-drag[1]) > 1)
		{
			if (drag[2] < drag[0]) std::swap(drag[2], drag[0]);
			if (drag[3] < drag[1]) std::swap(drag[3], drag[1]);

			if (drag[0] >= 0 && drag[1] >= 0 && drag[2] < mandelbrot->Width() && drag[3] < mandelbrot->Height())
			{
				Real ul[2];
				Real br[2];
				ul[0] = center[0] - scale[0] + 2.0 * scale[0] * Real(drag[0])/mandelbrot->Width();
				ul[1] = center[1] - scale[1] + 2.0 * scale[1] * Real(drag[1])/mandelbrot->Height();
				br[0] = center[0] - scale[0] + 2.0 * scale[0] * Real(drag[2])/mandelbrot->Width();
				br[1] = center[1] - scale[1] + 2.0 * scale[1] * Real(drag[3])/mandelbrot->Height();

				center[0] = (ul[0]+br[0])/2.0;
				center[1] = (ul[1]+br[1])/2.0;
				scale[0] = (br[0]-ul[0])/2.0;
				scale[1] = (br[1]-ul[1])/2.0;
				bChanged = true;
			}
		}
		// Cancel drag if no mouse left held
		if (!input.Get(MouseLeft))
			bDrag = false;

		// If the user changed something, reset everything and set up new starting conditions for a new calculation.
		if (bChanged)
			ResetMandelbrotArrays(center[0] - scale[0], center[1] - scale[1], center[0] + scale[0], center[1] + scale[1]);

		// Update calculations as many times as we can in the set period.
		static const float cUpdateTime = 20.f;
		float stopTime = Clock::Tick() + cUpdateTime * Clock::TicksPerSec() / 1000.f;
		int updates = 0;
		dword start = Clock::Tick();
		while(Clock::Tick() < stopTime && (numConvergingPoints > 0 || iterations == 0) && updates < cMaxIterationsPerFrame)
			updates += IterateFractal();
		
		iterationsPerSec = float(updates) * Clock::TicksPerSec() / (Clock::Tick()-start);

		// Convert the calculated values to a colored texture.
		DrawMandelbrot(*mandelbrot);
	}

	/// Renders using the GPU.
	void Render()
	{
		using namespace clb::Global;

		// Clear screen to black color (0) and set Z to max. far value (1.f)
		graphics.Clear(0, 1.f);

		// Render the mandelbrot image
		Rect dst(cTexX, cTexY, mandelbrot->Width(), mandelbrot->Height());
		graphics.RenderSprite(0, &dst, *mandelbrot);

		// Render info text
		Font *f = resPool.Get<Font>("systemFont");
		char text[1024];
		if (bUniformScale)
			sprintf_s(text, 1024, "Center: (%g,%g) Zoom: %gx Iterations: %d",
				center[0], center[1], 1.3/scale[0], iterations);
		else
			sprintf_s(text, 1024, "Center: (%g,%g) Zoom: (%gx,%gx) Iterations: %d",
				center[0], center[1], 1.3/scale[0], 1.3/scale[1],iterations);
		graphics.RenderText(*f, 70, 3, text);

		sprintf_s(text, 1024, "Converging: %d (%2.1f%%) Iterations per sec: %f", numConvergingPoints,
			numConvergingPoints*100.f/(cTextureSize*cTextureSize), iterationsPerSec);
		graphics.RenderText(*f, 70, 13, text);

		sprintf_s(text, 1024, "Arrows: Translate, Mouse Left Drag: Zoom in, Mouse Right: Zoom out, R: Reset");
		graphics.RenderText(*f, 70, graphics.Height() - 20, text, 0xFFFF8080);

		int x = input.Get(MouseX) - cTexX;
		int y = input.Get(MouseY) - cTexY;
		if (x >= 0 && y >= 0 && x < mandelbrot->Width() && y < mandelbrot->Height())
		{
			const Real xPos = center[0] - scale[0] + 2.0 * scale[0] * Real(x)/mandelbrot->Width();
			const Real yPos = center[1] - scale[1] + 2.0 * scale[1] * Real(y)/mandelbrot->Height();

			const Real &xVal = values[y * mandelbrot->Width() + x].value.r;
			const Real &yVal = values[y * mandelbrot->Width() + x].value.i;
			sprintf_s(text, 1024, "c=(%g,%g) -> z=(%g,%g) |z|=%g",
				(double)xPos, (double)yPos, (double)xVal, (double)yVal, sqrt((double)(xVal*xVal+yVal*yVal)));
			graphics.RenderText(*f, 70, 23, text);
		}
	}
};

clb::ClbLibApp *CreateApplicationObject()
{
	return new MandelbrotApp;
}